This patch add a function to exchange a foreign domain's one page to another.

Although there is already has a memory_exchange hypercall, but that hypercall will not accept new page, add that support will break current ABI. Also it does not support foreign domain. This function add those support.

Signed-off-by: Jiang, Yunhong <yunhong.jiang@intel.com>

diff -r 5a3bc2ebbc64 xen/arch/x86/mm.c
--- a/xen/arch/x86/mm.c	Thu Mar 19 06:14:15 2009 +0800
+++ b/xen/arch/x86/mm.c	Thu Mar 19 06:58:20 2009 +0800
@@ -2929,6 +2929,57 @@ int do_mmuext_op(
             put_page(mfn_to_page(src_mfn));
             break;
         }
+        case MMUEXT_EXCHANGE_PAGE:
+        {
+            unsigned long old_mfn, new_mfn;
+            xen_pfn_t old_pfn;
+
+            old_mfn = op.arg2.src_mfn;
+            new_mfn = op.arg1.mfn;
+
+            if ( !get_page_from_pagenr(new_mfn, FOREIGNDOM) )
+            {
+                okay = 0;
+                break;
+            }
+
+            if ( !FOREIGNDOM->is_shut_down ||
+                (FOREIGNDOM->shutdown_code != SHUTDOWN_suspend) )
+            {
+                okay = 0;
+                goto exchange_fail;
+            }
+
+            /* No referece to this page from page table anymore */
+
+            old_pfn = mfn_to_gmfn(FOREIGNDOM, old_mfn);
+
+            /* release original one */
+            if ( unlikely(steal_page(FOREIGNDOM, mfn_to_page(old_mfn), 0)) )
+            {
+                okay = 0;
+                goto exchange_fail;
+            }
+
+            if ( !test_and_clear_bit(_PGC_allocated,
+                  &(mfn_to_page(old_mfn)->count_info)) )
+                BUG();
+
+            guest_physmap_remove_page(FOREIGNDOM, old_pfn, old_mfn, 0);
+
+            put_page(mfn_to_page(old_mfn));
+
+            /* Setup the new page */
+            guest_physmap_add_page(FOREIGNDOM, old_pfn, new_mfn, 0);
+
+            if ( !paging_mode_translate(FOREIGNDOM) )
+            {
+                set_gpfn_from_mfn(new_mfn, old_pfn);
+            }
+exchange_fail:
+            put_page(mfn_to_page(new_mfn));
+            break;
+          }
 
         default:
             MEM_LOG("Invalid extended pt command 0x%x", op.cmd);
diff -r 5a3bc2ebbc64 xen/include/public/xen.h
--- a/xen/include/public/xen.h	Thu Mar 19 06:14:15 2009 +0800
+++ b/xen/include/public/xen.h	Thu Mar 19 06:16:22 2009 +0800
@@ -238,6 +238,10 @@ DEFINE_XEN_GUEST_HANDLE(xen_pfn_t);
  * cmd: MMUEXT_COPY_PAGE
  * mfn: Machine frame number of the destination page.
  * src_mfn: Machine frame number of the source page.
+ *
+ * cmd: MMUEXT_EXCHANGE_PAGE
+ * src_mfn: Machine frame number of the page that will be replaced by new_mfn
+ * mfn: the mfn of the new page.
  */
 #define MMUEXT_PIN_L1_TABLE      0
 #define MMUEXT_PIN_L2_TABLE      1
@@ -256,6 +260,7 @@ DEFINE_XEN_GUEST_HANDLE(xen_pfn_t);
 #define MMUEXT_NEW_USER_BASEPTR 15
 #define MMUEXT_CLEAR_PAGE       16
 #define MMUEXT_COPY_PAGE        17
+#define MMUEXT_EXCHANGE_PAGE        18
 
 #ifndef __ASSEMBLY__
 struct mmuext_op {
